from typing import List

import numpy as np

import matplotlib.pyplot as plt

from sklearn import preprocessing
from sklearn.datasets import make_classification


class PassiveAggressiveClassifier:
    def __init__(self, dimensionality):
        self.w = np.zeros(dimensionality)

    def value(self, x):
        return np.dot(x, self.w)

    def predict(self, x):
        return 1 if self.value(x) > 0 else -1

    def train(self, x, y):
        loss = min(1, max(0, 1 - y * self.value(x)))
        self.w += y * loss * x


def main():
    rng = np.random.RandomState()
    stream = DataStream(rng)
    dataset = stream.get_dataset()

    scores = []

    clf = PassiveAggressiveClassifier(stream.get_dimensionality())

    while stream.has_next():
        d = stream.get_next_data()

        # preprocessing: unit norm
        d = preprocessing.normalize(d.reshape(1, -1))[0]

        prediction = clf.predict(d)
        y = stream.get_actual_value()
        scores.append(score(clf, dataset))

        clf.train(d, y)

    visualize(dataset, scores)


def score(clf: PassiveAggressiveClassifier, dataset) -> float:
    xs, ys = dataset
    correct = 0
    for x, y in zip(xs, ys):
        if clf.predict(x) == y:
            correct += 1

    return correct / len(xs)


def visualize(dataset, scores: List[float]):
    xs, ys = dataset
    plt.subplot(211)
    plt.scatter(xs[:, 0], xs[:, 1], marker='o', c=ys, s=25)

    errors = [1-s for s in scores]
    plt.subplot(212)
    plt.plot(errors)
    plt.grid(True)
    plt.xlabel('Datapoint')
    plt.ylabel('Error')
    plt.ylim(ymin=0)
    plt.show()


class DataStream:
    def __init__(self, rng):
        X, y = make_classification(n_features=2, n_redundant=0, n_informative=2, n_clusters_per_class=1, random_state=rng)
        self.data = X
        self.target = [1 if v > 0 else -1 for v in y]
        self.index = -1

    def get_dimensionality(self):
        return len(self.data[0])

    def has_next(self):
        return self.index < len(self.data) - 1

    def get_next_data(self):
        self.index += 1
        return self.data[self.index]

    def get_actual_value(self):
        return self.target[self.index]

    def get_dataset(self):
        return self.data, self.target


if __name__ == '__main__':
    main()
